import os
from gwel.network import Segmenter
import json
import numpy as np
import cv2
from pycocotools import mask as mask_utils
from pycocotools.coco import COCO
from tqdm import tqdm

try:
    import joblib
    from sklearn.linear_model import LogisticRegression
    from sklearn.kernel_approximation import RBFSampler
    from sklearn.preprocessing import StandardScaler
    from sklearn.pipeline import Pipeline
except ImportError:
    import subprocess
    import sys
    subprocess.check_call([sys.executable, "-m", "pip", "install", "scikit-learn","joblib"])
    from sklearn.linear_model import LogisticRegression
    from sklearn.kernel_approximation import RBFSampler
    from sklearn.preprocessing import StandardScaler
    from sklearn.pipeline import Pipeline
    import joblib

class GLMSegmenter(Segmenter):
    def __init__(self, model_path: str, class_names: list ):
        self.model_path = model_path
        self.load_weights(model_path)
        self.channels = class_names

    def set_device(self, device: str):
        pass

    def load_weights(self, weights: str = None):
        path = weights or self.model_path
        self.model = joblib.load(path)

    def inference(self, image: np.ndarray):

        H, W, _ = image.shape
        image_rgb = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        pixels = image_rgb.reshape(-1, 3)

        y_pred = self.model.predict(pixels)          # shape [H*W]
       # y_proba = self.model.predict_proba(pixels)   # shape [H*W, C]

        results = {}
        for i, class_name in enumerate(self.channels):
            mask = (y_pred == i).astype(np.uint8).reshape(H, W)
            rle = mask_utils.encode(np.asfortranarray(mask))
            results[class_name] = rle

        return results



def load_annotations(coco_json_path):
    return COCO(coco_json_path)

def load_mask(coco, ann_ids, height, width):
    anns = coco.loadAnns(ann_ids)
    mask = np.zeros((height, width), dtype=np.uint8)
    for ann in anns:
        rle = mask_utils.frPyObjects(ann['segmentation'], height, width)
        rle = mask_utils.merge(rle)
        m = mask_utils.decode(rle)
        mask = np.logical_or(mask, m)
    return mask.astype(np.uint8)

def sample_pixels(image, mask, num_samples_per_class=500):
    """
    image: HxWx3 RGB
    mask:  HxW binary (1 for leaf, 0 for background)
    """
    H, W = mask.shape
    coords_leaf = np.argwhere(mask == 1)
    coords_bg   = np.argwhere(mask == 0)

    if len(coords_leaf) == 0 or len(coords_bg) == 0:
        return [], []

    coords_leaf = coords_leaf[np.random.choice(len(coords_leaf), min(num_samples_per_class, len(coords_leaf)), replace=False)]
    coords_bg   = coords_bg[np.random.choice(len(coords_bg), min(num_samples_per_class, len(coords_bg)), replace=False)]

    samples = np.vstack([coords_leaf, coords_bg])
    labels  = np.array([1] * len(coords_leaf) + [0] * len(coords_bg))

    rgb = image[samples[:,0], samples[:,1], :]
    return rgb, labels

def train_GLM_on_coco(coco_json, image_dir, output_model_path, samples_per_image=1000):
    coco = load_annotations(coco_json)
    cat_ids = coco.getCatIds(catNms=['leaf'])
    img_ids = coco.getImgIds()

    X_all = []
    y_all = []

    for img_id in tqdm(img_ids):
        img_info = coco.loadImgs([img_id])[0]
        ann_ids = coco.getAnnIds(imgIds=[img_id], catIds=cat_ids)
        if not ann_ids:
            continue

        image_path = os.path.join(image_dir, img_info['file_name'])
        if not os.path.exists(image_path):
            continue

        image = cv2.imread(image_path)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        mask = load_mask(coco, ann_ids, img_info['height'], img_info['width'])

        X, y = sample_pixels(image, mask, num_samples_per_class=samples_per_image // 2)
        if len(X) == 0:
            continue

        X_all.append(X)
        y_all.append(y)

    X_all = np.vstack(X_all)
    y_all = np.concatenate(y_all)

    print(f"Total samples: {len(X_all)}")

    # Define kernelized softmax model
    model = Pipeline([
        ('scale', StandardScaler()),
        ('rbf', RBFSampler(gamma=0.5, n_components=10)),
        ('logreg', LogisticRegression(multi_class='multinomial', solver='lbfgs', max_iter=500))
    ])

    model.fit(X_all, y_all)
    joblib.dump(model, output_model_path)
    print(f"✅ Model saved to: {output_model_path}")

def train_GLM_on_palettes(classes = [], image_paths = [], output_model_path, samples_per_image=1000):

    X_all = []
    y_all = []

    if len(classes) != len(image_paths):
        

    for image in image_paths:
        img_info = coco.loadImgs([img_id])[0]
        ann_ids = coco.getAnnIds(imgIds=[img_id], catIds=cat_ids)
        if not ann_ids:
            continue

        image_path = os.path.join(image_dir, img_info['file_name'])
        if not os.path.exists(image_path):
            continue

        image = cv2.imread(image_path)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        mask = load_mask(coco, ann_ids, img_info['height'], img_info['width'])

        X, y = sample_pixels(image, mask, num_samples_per_class=samples_per_image // 2)
        if len(X) == 0:
            continue

        X_all.append(X)
        y_all.append(y)


    image = cv2.imread(image_path)
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    mask = np.ones_like(image)

    X, y = sample_pixels(image, mask, num_samples_per_class=samples_per_image)
    if len(X) == 0:
        continue

    X_all.append(X)
    y_all.append(y)

    X_all = np.vstack(X_all)
    y_all = np.concatenate(y_all)

    print(f"Total samples: {len(X_all)}")

    # Define kernelized softmax model
    model = Pipeline([
        ('scale', StandardScaler()),
        ('rbf', RBFSampler(gamma=0.5, n_components=10)),
        ('logreg', LogisticRegression(multi_class='multinomial', solver='lbfgs', max_iter=500))
    ])

    model.fit(X_all, y_all)
    joblib.dump(model, output_model_path)
    print(f" Model saved to: {output_model_path}")



from gwel.network import Segmenter
import torch
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import cv2
from pycocotools import mask as mask_utils

class UNet(nn.Module):
    def __init__(self, in_channels, out_channels):
        super(UNet, self).__init__()

        # Contracting path (Encoder)
        self.encoder1 = self.conv_block(in_channels, 64)
        self.encoder2 = self.conv_block(64, 128)
        self.encoder3 = self.conv_block(128, 256)
        self.encoder4 = self.conv_block(256, 512)

        # Expanding path (Decoder)
        self.decoder4 = self.upconv_block(512, 256)
        self.decoder3 = self.upconv_block(256, 128)
        self.decoder2 = self.upconv_block(128, 64)
        self.decoder1 = self.out_layer(64, out_channels)


    def conv_block(self, in_channels, out_channels):
        return nn.Sequential(
            nn.Conv2d(in_channels, out_channels, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(out_channels, out_channels, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=4,stride=2,padding=1)
        )

    def upconv_block(self, in_channels, out_channels):
        return nn.Sequential(
            nn.ConvTranspose2d(in_channels, out_channels, kernel_size=4, stride=2,padding=1),
            nn.ReLU(inplace=True),
	    nn.Conv2d(out_channels,out_channels,kernel_size=3,padding=1),
            nn.ReLU(inplace=True)
        )
    
    def out_layer(self,in_channels,out_channels):
        return nn.Sequential(
	    nn.ConvTranspose2d(in_channels,out_channels,kernel_size=4,stride = 2,padding=1),
	    nn.Conv2d(out_channels,out_channels,kernel_size=3,padding=1)
        )

    def forward(self, x):
        # Encoder
        e1 = self.encoder1(x)
        e2 = self.encoder2(e1)
        e3 = self.encoder3(e2)
        e4 = self.encoder4(e3)

        # Decoder
        d3 = self.decoder4(e4)
        d2 = self.decoder3(d3 + e3)  
        d1 = self.decoder2(d2 + e2)   
        out = self.decoder1(d1 + e1)   
        return out

class CenterPad:
    def __init__(self, target_multiple=256):
       self.target_multiple = target_multiple
    
    def __call__(self, tensor):

        if len(tensor.shape) == 3:  
            _, height, width = tensor.shape
        elif len(tensor.shape) == 2: 
            height, width = tensor.shape
            
            tensor = tensor.unsqueeze(0)  
        else:
            raise ValueError("Input tensor must have 2 or 3 dimensions.")
        

        _, height, width = tensor.shape 

        pad_h = (self.target_multiple - height % self.target_multiple) % self.target_multiple
        pad_w = (self.target_multiple - width % self.target_multiple) % self.target_multiple
        
        pad_h_top = pad_h // 2
        pad_h_bottom = pad_h - pad_h_top
        pad_w_left = pad_w // 2
        pad_w_right = pad_w - pad_w_left
        
        padded_tensor = torch.nn.functional.pad(tensor, (pad_w_left, pad_w_right, pad_h_top, pad_h_bottom),
                                                value=0)
        return padded_tensor



class UNET(Segmenter):
    def __init__(self,
                 weights : str ,
                 patch_size : int,
                 channels : list):
        
        self.model = UNet(3,len(channels)+1)
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu") 
        self.set_device(self.device)
        self.weights = weights
        self.load_weights( self.weights)
        self.patch_size = patch_size
        self.centerpad = CenterPad(patch_size)
        self.channels = channels

    def set_device(self, device : str):
        self.device = device
        self.model.to(device)

    def load_weights(self, weights : str = None):
        self.model.load_state_dict(torch.load(weights, map_location=self.device,weights_only=True)['model_state_dict'])
        self.model.eval()


    def inference(self, image : np.ndarray):
 
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        image = torch.as_tensor(image,dtype=torch.float).permute(2,0,1) / 255
        image_padded = self.centerpad(image)
        image_padded = image_padded.to(self.device)
        
        with torch.no_grad():
            output = self.model(image_padded)
            #output = torch.sigmoid(output)
            output =  torch.argmax(output,dim=0)
        output = output.cpu().numpy()


        _, crop_height, crop_width = image.shape
        height, width = output.shape
        left = (width - crop_width) // 2
        top = (height - crop_height) // 2
        right = (width + crop_width) // 2
        bottom = (height + crop_height) // 2 
        
        cropped_masks = output[top:bottom, left:right]
        
        output = {}
        
        for channel in range(len(self.channels)+1):
            mask = (cropped_masks == channel).astype(np.uint8)  # ensure mask is uint8

            # Apply morphological closing
            #kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (3, 3))
            #mask = cv2.morphologyEx(mask, cv2.MORPH_CLOSE, kernel)

            rle = mask_utils.encode(np.asfortranarray(mask))

            if channel == 0:
                output['background'] = rle
            else:
                output[self.channels[channel-1]] = rle

        return output



import numpy as np
import cv2
import warnings
import os
from gwel.network import Detector
import subprocess
import sys
import torch
from torchvision.ops import nms


import sys

try:
    from ultralytics import YOLO
except ImportError:
    sys.exit("Ultralytics not found. Install with 'pip install ultralytics'.")



warnings.filterwarnings("ignore")

def bbox_to_polygon(bboxes):
    polygons = []
    for bbox in bboxes:
        x1, y1, x2, y2 = bbox
        polygon = [[x1, y1], [x1, y2], [x2, y2], [x2, y1]]
        polygons.append([np.array(polygon,dtype=np.int32).reshape(-1, 1, 2)])
    return polygons

class YOLOv8(Detector):

    def __init__(self, weights: str , device: str = "cpu", patch_size: tuple = None):
        self.threshold = 0.5
        self.patch_size = patch_size
        self.device = device
        if weights:
            self.load_weights(weights)
        #self.set_device(device)

    def set_device(self, device: str):
        self.device = device
        if hasattr(self, 'model'):
            self.model.to(self.device)

    def load_weights(self, weights: str):
        self.weights = weights
        self.model = YOLO(weights, task = 'detect')
        if torch.cuda.is_available():
            device = torch.device('cuda')
            self.model.to(device)
        #self.model.to(self.device)
     
    def inference(self, image: np.ndarray):
        if not self.patch_size:
            results = self.model.predict(image,verbose=False, device = self.device)
            boxes = results[0].boxes.xyxy.cpu().numpy() 
            detections = boxes.tolist()
            detections = bbox_to_polygon(detections)
        else:
            detections = self.inference_with_patches(patch_size = self.patch_size, image = image)
        return detections
    """
    def inference_with_patches(self, patch_size: tuple, image: np.ndarray):
        h, w = image.shape[:2]
        patch_h, patch_w = patch_size
        pad_h = (patch_h - h % patch_h) % patch_h
        pad_w = (patch_w - w % patch_w) % patch_w
        
        padded_image = cv2.copyMakeBorder(image, 0, pad_h, 0, pad_w, cv2.BORDER_CONSTANT, value=(0, 0, 0))
        detections = []

        for i in range(0, padded_image.shape[0], patch_h):
            for j in range(0, padded_image.shape[1], patch_w):
                patch = padded_image[i:i + patch_h, j:j + patch_w] 
                results = self.model.predict(patch, verbose=False)
                boxes = results[0].boxes.xyxy.cpu().numpy()
                for bbox in boxes:
                    x1, y1, x2, y2 = bbox
                    x1 += j 
                    y1 += i  
                    x2 += j
                    y2 += i

                    x1, y1 = max(0, x1), max(0, y1)
                    x2, y2 = min(w, x2), min(h, y2)

                    detections.append([x1, y1, x2, y2])


        return bbox_to_polygon(detections)
    """
    def inference_with_patches(
        self,
        patch_size: tuple,
        image: np.ndarray,
        overlap_ratio: float = 0.2,
        iou_thresh: float = 0.2,
        border_margin: float = 0.05,
    ):
        h, w = image.shape[:2]
        patch_h, patch_w = patch_size

        stride_h = int(patch_h * (1 - overlap_ratio))
        stride_w = int(patch_w * (1 - overlap_ratio))

        pad_h = (patch_h - h % stride_h) % stride_h
        pad_w = (patch_w - w % stride_w) % stride_w

        padded_image = cv2.copyMakeBorder(
            image, 0, pad_h, 0, pad_w,
            cv2.BORDER_CONSTANT, value=(0, 0, 0)
        )

        detections = []

        # overlap-aware border margin
        margin_x = int(patch_w * overlap_ratio * border_margin)
        margin_y = int(patch_h * overlap_ratio * border_margin)

        for i in range(0, padded_image.shape[0] - patch_h + 1, stride_h):
            for j in range(0, padded_image.shape[1] - patch_w + 1, stride_w):
                patch = padded_image[i:i + patch_h, j:j + patch_w]
                results = self.model.predict(patch, verbose=False)

                boxes = results[0].boxes.xyxy.cpu().numpy()
                confs = results[0].boxes.conf.cpu().numpy()

                
                for bbox, conf in zip(boxes, confs):
                    x1, y1, x2, y2 = bbox

                    x1 += j
                    y1 += i
                    x2 += j
                    y2 += i

                    x1, y1 = max(0, x1), max(0, y1)
                    x2, y2 = min(w, x2), min(h, y2)

                    # discard boxes touching slice border
                    if (
                        x1 <= j + margin_x or
                        y1 <= i + margin_y or
                        x2 >= j + patch_w - margin_x or
                        y2 >= i + patch_h - margin_y
                    ):
                        continue

                    detections.append([x1, y1, x2, y2, conf])


        if len(detections) == 0:
            return np.empty((0, 4))

        detections = torch.tensor(detections, dtype=torch.float32)
        boxes = detections[:, :4]
        scores = detections[:, 4]

        keep_idx = nms(boxes, scores, iou_thresh)

        detections = boxes[keep_idx].cpu().numpy()

        return bbox_to_polygon(detections)




# src/gwel/cli.py
import os
import json
import typer
from typing import List
from gwel.dataset import ImageDataset
from gwel.viewer import Viewer



app = typer.Typer(add_completion = True, help="GWEL CLI tool", invoke_without_command=True, context_settings={"help_option_names": ["-h", "--help"]})

try:
    from importlib.metadata import version, PackageNotFoundError
except ImportError:
    from importlib_metadata import version, PackageNotFoundError  # For Python <3.8

try:
    VERSION = version("gwel")
except PackageNotFoundError:
    VERSION = "unknown"



@app.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    version_flag: bool = typer.Option(
        False, "--version", "-v", help="Show GWEL CLI version"
    ),
):
    """
    Callback function that runs when 'gwel' is called without a subcommand.
    """
    if version_flag:
        typer.secho(f"GWEL CLI version {VERSION}", fg=typer.colors.GREEN, bold=True)
        raise typer.Exit()

    if ctx.invoked_subcommand is None:
        typer.secho(f"GWEL CLI version {VERSION}", fg=typer.colors.CYAN, bold=True)
        typer.echo("Available commands:")

        # Use rich-like table formatting with Typer
        for name, cmd in ctx.command.commands.items():
            description = cmd.help or cmd.callback.__doc__ or ""
            if description:
                description = description.strip().split("\n")[0]
            typer.secho(f"  {name:<10}", fg=typer.colors.YELLOW, bold=True, nl=False)
            typer.echo(f" {description}")




@app.command()
def view(
    ctx: typer.Context,
    resized_flag: bool = typer.Option(
        False, "--resized", "-r", help="View resized images."),
     detections_flag: bool = typer.Option(
        False, "--detections", "-d", help="Load pre-detected object annotations."), 
    segmentation_flag: bool = typer.Option(
        False, "--segmentation", "-s", help="Load pre-detected segementation."),  
    max_pixels: int = typer.Option(
        800, "--maxpixels", "-p", help="Max Pixels."),
    index: int = typer.Option(
        1, "--index", "-i", help="The index of image to be viewed on opening."),
    contour_thickness: int = typer.Option(
        2, "--thickness", "-t", help="Thickness in pixels of annotations."), 
    col_scheme: str = typer.Option(
        {}, "--col_scheme", "-c", help="Color scheme for annotations as JSON string using color names recognized by matplotlib." 
         )
):
    """
    Open the image viewer with images from the current directory.
    """
    directory = os.getcwd()
    try:
        dataset = ImageDataset(directory)
        if resized_flag:
            dataset.resize()
        if detections_flag:
            dataset.detect()
        if segmentation_flag:
            dataset.segment() 
        col_scheme_dict = json.loads(col_scheme) 
        viewer = Viewer(dataset,max_pixels = max_pixels,contour_thickness=contour_thickness,col_scheme=col_scheme_dict)
        viewer.index = index - 1
        if detections_flag:
            viewer.mode = 'instance'
        viewer.open()
    except ValueError as e:
        # Only print the error message, no traceback
        typer.secho(f"Error: {e}", fg=typer.colors.RED, bold=True)



@app.command()
def resize(max_pixels: int = typer.Option(
        800, "--maxpixels", "-p", help="Max Pixels.")):
    """
    Create resized copies of the images from the current directory.
    """
    directory = os.getcwd()
    try:
        dataset = ImageDataset(directory)
        dataset.resize(max_size=max_pixels)
    except ValueError as e:
        # Only print the error message, no traceback
        typer.secho(f"Error: {e}", fg=typer.colors.RED, bold=True)



@app.command()
def detect(model: str = typer.Argument(...,help="Model type"), 
           weights: str = typer.Argument(...,help="Path to model weights"), 
           slice_size:int = typer.Option(
    None, "--slicesz", "-s", help="Slice size")):
    """
    Run a detector on the images from the current directory.
    """
    directory = os.getcwd()
    try:
        if os.path.exists(weights):
            if model == "YOLOv8":
                from gwel.networks.YOLOv8 import YOLOv8
                if not slice_size:
                    detector = YOLOv8(weights)
                else:
                    detector = YOLOv8(weights,patch_size=(slice_size,slice_size))
            else:
                raise ValueError("Model type unknown.")
        else:
            raise ValueError("No weights found at location {weights}.")
        dataset = ImageDataset(directory)
        dataset.resize()
        dataset.detect(detector)
    except ValueError as e:
        # Only print the error message, no traceback
        typer.secho(f"Error: {e}", fg=typer.colors.RED, bold=True)




@app.command()
def segment(model: str = typer.Argument(...,help="Model type"),
            weights: str = typer.Argument(...,help="Path to model weights"), 
            channels: List[str] = typer.Argument(...,help="Segmentation class labels"), 
            patch_size:int = typer.Option(256, "--patchsz", "-s", help="Patch size"),
            background:bool = typer.Option(False, "--background","-b", help="Include background channel")
            ):
    """
    Run a segmenter on the images from the current directory.
    """
    directory = os.getcwd()
    try:
        if os.path.exists(weights):
            if model == "UNET":
                from gwel.networks.UNET import UNET  
                segmenter = UNET(weights, patch_size, channels)
            else:
                raise ValueError("Model type unknown.")
        else:
            raise ValueError("No weights found at location {weights}.")
        dataset = ImageDataset(directory)
        if background:
            dataset.segment(segmenter, background=True)
        else:   
            dataset.segment(segmenter)

    except ValueError as e:
        # Only print the error message, no traceback
        typer.secho(f"Error: {e}", fg=typer.colors.RED, bold=True)

@app.command()
def crop(path: str, object_name : str):
    """
    Crop the images from the current directory based on the bounding boxes of detections.
    """
    directory = os.getcwd()
    try:
        dataset = ImageDataset(directory)
        dataset.detect()
        dataset.crop(path,object_name)
    except ValueError as e:
        # Only print the error message, no traceback
        typer.secho(f"Error: {e}", fg=typer.colors.RED, bold=True)


if __name__ == "__main__":
    app()




import os
import json
import numpy as np
import torch
from torch.utils.data import Dataset, DataLoader
from PIL import Image
from pycocotools.coco import COCO
from torchvision import transforms

class CenterPad:
    def __init__(self, target_multiple=256):
       self.target_multiple = target_multiple
    
    def __call__(self, tensor):

        if len(tensor.shape) == 3:  
            _, height, width = tensor.shape
        elif len(tensor.shape) == 2: 
            height, width = tensor.shape
            
            tensor = tensor.unsqueeze(0)  
        else:
            raise ValueError("Input tensor must have 2 or 3 dimensions.")
        

        _, height, width = tensor.shape  # Assuming tensor has shape (C, H, W)
        
        pad_h = (self.target_multiple - height % self.target_multiple) % self.target_multiple
        pad_w = (self.target_multiple - width % self.target_multiple) % self.target_multiple
        
        pad_h_top = pad_h // 2
        pad_h_bottom = pad_h - pad_h_top
        pad_w_left = pad_w // 2
        pad_w_right = pad_w - pad_w_left
        
        padded_tensor = torch.nn.functional.pad(tensor, (pad_w_left, pad_w_right, pad_h_top, pad_h_bottom),
                                                value=0)
        return padded_tensor


class CocoSegmentationDataset(Dataset):
    def __init__(self, image_dir, coco_json, categories: list, transform = transforms.Compose([CenterPad(target_multiple=256)])):
        self.coco = COCO(coco_json)
        self.image_dir = image_dir
        self.image_ids = self.coco.getImgIds()
        self.categories = categories
        self.transform = transform

    def __len__(self):
        return len(self.image_ids)

    def __getitem__(self, idx):
        img_id = self.image_ids[idx]
        img_info = self.coco.loadImgs([img_id])[0]
        img_path = os.path.join(self.image_dir, img_info['file_name'])
        image = Image.open(img_path).convert("RGB")

        # Target mask initialized as background = 0
        target = np.zeros((image.height, image.width), dtype=np.uint8)

        # Loop over categories
        for catIdx, _ in enumerate(self.categories):  
            ann_ids = self.coco.getAnnIds(imgIds=[img_id], catIds=[catIdx+1])
            annotations = self.coco.loadAnns(ann_ids)

            mask = np.zeros((image.height, image.width), dtype=np.uint8)
            for ann in annotations:
                ann_mask = self.coco.annToMask(ann)
                mask = np.maximum(mask, ann_mask)

            target[mask == 1] = catIdx + 1  # assign class index

        # Convert to tensors
        image = torch.as_tensor(np.array(image), dtype=torch.float32).permute(2, 0, 1) / 255.0
        target = torch.as_tensor(target, dtype=torch.long)

        # Apply transforms 
        if self.transform is not None:
            image = self.transform(image)
            target = self.transform(target)
            target = target.squeeze(0).long()
        
        return image, target




import numpy as np
import cv2
import warnings
warnings.filterwarnings("ignore")

from detectron2.engine import DefaultPredictor
from detectron2.config import get_cfg
from detectron2.projects.deeplab import add_deeplab_config
from detectron2.structures import BitMasks

from mask2former import add_maskformer2_config

from gwel.network import Detector


class Mask2Former(Detector):

    def __init__(self, arch_yaml: str, weights: str = None, device : str = "cpu"):
        self.cfg = get_cfg()
        self.cfg.set_new_allowed(True)
        add_deeplab_config(self.cfg)
        add_maskformer2_config(self.cfg)
        self.cfg.merge_from_file(arch_yaml)
        self.threshold = 0.9

        if weights:
            self.load_weights(weights)
    
        self.set_device(device = device)

    def set_device(self, device : str):
        self.cfg.MODEL.DEVICE = device

    def load_weights(self, weights : str):
        self.cfg.MODEL.WEIGHTS = weights


    def inference(self, image : np.ndarray):
        cfg = self.cfg
        cfg.MODEL.MASK_FORMER.TEST.INSTANCE_ON = True
        cfg.MODEL.ROI_HEADS.SCORE_THRESH_TEST = 0.7
        predictor = DefaultPredictor(cfg)

        outputs = predictor(image)

        instances = outputs["instances"].to("cpu")
        high_conf_indices = np.where(instances.scores >= self.threshold)[0]  # Only keep high confidence
        filtered_instances = instances[high_conf_indices]

        pred_masks = filtered_instances.pred_masks
        image_height, image_width = filtered_instances.image_size

        bitmasks = BitMasks(pred_masks)
        contours_list = []
        for mask in bitmasks.tensor.cpu().numpy():  # Convert tensor to NumPy array
            contours, _ = cv2.findContours(mask.astype("uint8"), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
            contours_list.append(contours)
        
        return contours_list




